#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

static const char* progname = NULL;

static const char* Basename(const char* fullpath)
{
  const char* rv = strrchr(fullpath, (int)'/');
  if( rv == NULL ) {
    rv = fullpath;
  } else {
    ++rv;
  }
  return rv;
}

static void Usage(int exit_value)
{
  FILE* f = stdout;
  if( exit_value ) {
    f = stderr;
  }
  fprintf(f, "Usage: %s [-q] [-u] <template>\n", progname);
  exit(exit_value);
}

static int my_mkstemp(char* tempFile)
{
  char* basename = NULL;
  char* testname = NULL;
  unsigned i = 0;
  int fd = -1;

  /* Verify that the last six characters of template are "XXXXXX". */
  unsigned len = strlen(tempFile);
  if( len < 6 ) {
    errno = EINVAL;
    return -1;
  }
  if( strncmp(tempFile + (len - 6), "XXXXXX", 6) != 0 ) {
    errno = EINVAL;
    return -1;
  }

  /*
   * Generate a unique file name and open exclusively.  Selecting a
   * file name is usually done randomly and with letters, but what
   * follows is good enough.
   */

  if( (basename = strdup(tempFile)) == NULL ) {
    fprintf(stderr, "%s: strdup: %s\n", progname, strerror(errno));
    exit(1);
  }
  basename[len - 6] = '\0';

  if( (testname = (char*)malloc(len + 1)) == NULL ) {
    fprintf(stderr, "%s: malloc: %s\n", progname, strerror(errno));
    exit(1);
  }

  for( i = 0 ; i <= 999999 ; ++i ) {
    snprintf(testname, len + 1, "%s%06d", basename, i);
    fd = open(testname, (O_RDWR | O_CREAT | O_EXCL), (S_IRUSR | S_IWUSR));
    if( fd > 0 ) {
      strncpy(tempFile, testname, len);
      break;
    }
  }

  if( testname ) {
    free(testname);
    testname = NULL;
  }

  if( basename ) {
    free(basename);
    basename = NULL;
  }

  if( i > 999999 ) {
    errno = EEXIST;
    return -1;
  }
    
  return fd;
}


int main(int argc, char* argv[])
{
  int quiet = 0;
  int unsafe = 0;
  char* tempFile = NULL;
  int i;

  progname = argv[0];

  /* Evaluate command-line parameters. */
  for( i = 1 ; i < argc ; ++i ) {
    char* param= argv[i];
    if( strncmp(param, "-q", 2) == 0 ) {
      quiet = 1;
    } else if( strncmp(param, "-u", 2) == 0 ) {
      unsafe = 1;
    } else {
      if( (tempFile = strdup(param)) == NULL ) {
        fprintf(stderr, "%s: strdup: %s\n", progname, strerror(errno));
        exit(1);
      }
      break;
    }
  }

  /* Verify command-line parameters. */
  if( (tempFile == NULL) || (i < argc - 1) ) {
    Usage(1);
    return 1;
  }

  /* Do it. */
  if( (my_mkstemp(tempFile) == -1) && !quiet ) {
    fprintf(stderr, "%s: mkstemp: %s\n", progname, strerror(errno));
    return 1;
  }

  /* Use of unsafe mode places your shell script in the same position
   * as a C program that uses tmpnam(). */
  if( unsafe ) {
    if( (remove(tempFile) == -1) && !quiet ) {
      fprintf(stderr, "%s: remove: %s\n", progname, strerror(errno));
      return 1;
    }
  }

  printf("%s\n", tempFile);

  if( tempFile ) {
    free(tempFile);
    tempFile = NULL;
  }

  return 0;
}
